# Egor Maximenko and Mario Alberto Moctezuma Salazar, 2017-05-28
# In this program we perform some symbolic computations
# to verify Theorems 2.1, 2.2, 2.3 from our paper
# "Cofactors and eigenvectors of banded Toeplitz matrices:
# Trench formulas via skew Schur polynomials".
# https://arxiv.org/abs/1705.08067

Sym = SymmetricFunctions(QQ)
schur = Sym.s()
elem = Sym.e()
hom = Sym.h()
import itertools
import sys

def componentwise_ge(sp):
   la, mu = sp
   return (len(la) >= len(mu)) and all([la[j] >= mu[j] for j in range(len(mu))])

def skew_schur(sp):
   return schur.skew_schur(SkewPartition(sp)) if componentwise_ge(sp) else 0 * schur()

def elemfun(j):
   return elem([j]) if (j >= 0) else 0

def hom_fun(j):
   return schur(hom([j])) if (j >= 0) else 0

def toeplitz_matrix(p, n):
   elems = [[((-1) ** (p - r + s)) * schur(elemfun(p - r + s)) for s in range(n)] for r in range(n)]
   return matrix(schur, n, n, elems)

# The Sage function "det" crashes for some symbolic matrices.
# The trivial recursive version works better.
# The sum over all permutations is a little slower because our matrices have many zeros.

def det_recur(A, rfirst, cols):
   if rfirst == A.nrows():
      return A.base_ring().one()
   result = A.base_ring().zero()
   s = 1
   for k in cols:
      if A[rfirst][k] != 0:
         newcols = [c for c in cols if c != k]
         result += A[rfirst][k] * det_recur(A, rfirst + 1, newcols) * s
      s = -s
   return result

def trivial_det(A):
   return det_recur(A, 0, list(range(A.nrows())))

def matrix_minor(A, rows, cols):
   return trivial_det(A[rows, cols])

def matrix_cofactor(A, r, s):
   n = A.nrows()
   rows = [j for j in range(n) if j != r]
   cols = [k for k in range(n) if k != s]
   sgn = (-1) ** (r + s)
   return sgn * matrix_minor(A, rows, cols)

def matrix_adjugate(A):
   n = A.nrows()
   elems = [[matrix_cofactor(A, s, r) for s in range(n)] for r in range(n)]
   return matrix(A.base_ring(), n, n, elems)

def toeplitz_minor(p, n, rows, cols):
   return matrix_minor(toeplitz_matrix(p, n), rows, cols)

def toeplitz_adjugate(p, n):
   return matrix_adjugate(toeplitz_matrix(p, n))

def complement(n, mylist):
   setmylist = set(mylist)
   return [j for j in range(n) if j not in setmylist]

def toeplitz_minor_struck_out(p, n, xi, eta):
   return toeplitz_minor(p, n, complement(n, xi), complement(n, eta))

def difrev(xi):
   return list(reversed([xi[j] - j for j in range(len(xi))]))

def our_minor_formula(p, n, xi, eta):
   la = [n - len(xi)] * p + difrev(xi)
   mu = difrev(eta)
   sgn = (-1) ** (p * (n - len(xi)) + sum(xi) + sum(eta))
   return 0 * schur() + sgn * skew_schur([la, mu])

def ale_minor_formula(p, n, xi, eta):
   d = len(xi)
   m = n - d
   al = [m] * p + [m + j - eta[j] for j in range(d)]
   be = [m + j - xi[j] for j in range(d)]
   sgn = (-1) ** (p * (n - len(xi)) + sum(xi) + sum(eta))
   return 0 * schur() + sgn * skew_schur([al, be])

def our_adjugate_formula(p, n, r, s):
   # follows from our formula for minors
   la = [n - 1] * p + [s]
   mu = [r]
   sgn = (-1) ** (p * (n - 1))
   return sgn * skew_schur([la, mu])

def ale_adjugate_formula(p, n, r, s):
   # follows from Alexandersson's formula
   al = [n - 1] * p + [n - r - 1]
   be = [n - s - 1]
   sgn = (-1) ** (p * (n - 1))
   return sgn * skew_schur([al, be])

def moc_adjugate_formula(p, n, r, s):
   # expansion in Schur functions found by Moctezuma Salazar
   kmin = max(0, s - r)
   kmax = min(n - r - 1, s)
   np = [n - 1] * (p - 1)
   sgn = (-1) ** (p * (n - 1))
   return sgn * sum(schur(np + [n + s - r - 1 - k, k]) for k in range(kmin, kmax + 1))

def tre_adjugate_formula(p, n, r, s):
   # Trench formula for Toeplitz adjugate matrix
   s1 = hom_fun(s - r - p) * schur([n] * p)
   s2 = sum(((-1) ** k) * hom_fun(s + 1 + k - p) * schur([n] * (p - k - 1) + [n - 1] * k + [n - r - 1]) for k in range(p))
   sgn = (-1) ** (p * n)
   return sgn * (s1 - s2)

def xlist(w):
   return elem([0]).expand(w).args()

def polring(w):
   return elem([0]).expand(w).parent()

def toeplitz_eigenvector(p, n):
   elems = [schur([n - 1] * (p - 1) + [n - r - 1]) for r in range(n)]
   return matrix(schur, n, 1, elems)

def matrix_from_schur_to_polynomial(A, w):
   m, n = A.nrows(), A.ncols()
   elems = [[A[j, k].expand(w) for k in range(n)] for j in range(m)]
   return matrix(polring(w), m, n, elems)

def schur_numerator_matrix(la, w):
   laext = la + [0] * (w - len(la))
   x = xlist(w)
   elems = [[x[k] ** (laext[j] + (w - 1 - j)) for k in range(w)] for j in range(w)]
   return matrix(polring(w), w, w, elems)

def vandermonde_polynomial(w):
   return trivial_det(schur_numerator_matrix([0], w))

# Let us compute the power of wpol[r]:
# the power of Gelems[r, k] is n-r-1+w-p, and
# the power of the cofactor is (n+w-1+...+n+w-p+1)+(w-p-1+...+0)=(2n+2w-p)*(p-1)/2
# the power of D is w-1+...+0=w(w-1)/2

def toeplitz_eigenvector_via_geometric_progressions(p, n, w):
   x = xlist(w)
   A = schur_numerator_matrix([n - 1] * p, w)
   D = vandermonde_polynomial(w)
   PR = polring(w)
   Gelems = [[x[k] ** (n - r - 1 + w - p) for k in range(w)] for r in range(n)]
   G = matrix(PR, n, w, Gelems)
   Celems = [matrix_cofactor(A, p - 1, k) for k in range(w)]
   C = matrix(PR, w, 1, Celems)
   GC = G * C
   wpol = GC / D
   return wpol


# Tests and interface:
# verbosity = 0: print nothing, return true or false,
# verbosity = 1: output the numbers of tests, output the parameters if some test is failed,
# verbosity = 2: output the parameters of each individual test,
# verbosity = 3: output the object computed by various methods.

def pr(s):
   sys.stdout.write(s)
   sys.stdout.flush()

def verify_minor(p, n, xi, eta, verbosity = 0):
   if verbosity >= 2:
      pr('Verify_minor with p = {0}, n = {1}, xi = {2}, eta = {3} ... '.format(p, n, xi, eta))
   minor = toeplitz_minor_struck_out(p, n, xi, eta)
   ourformula = our_minor_formula(p, n, xi, eta)
   aleformula = ale_minor_formula(p, n, xi, eta)
   if verbosity >= 3:
      pr('Minor      = {0}\n'.format(minor))
      pr('OurFormula = {0}\n'.format(ourformula))
      pr('AleFormula = {0}\n'.format(aleformula))
   if minor == ourformula and ourformula == aleformula:
      if verbosity >= 2:
         pr('correct.\n')
      return True
   else:
      if verbosity >= 1:
         pr('\nThe minors do not coincide with p = {0}, n = {1}, xi = {2}, eta = {3}.\n'.format(p, n, xi, eta))
      return False

def full_test_minors(pmax, nmax, verbosity = 0):
   if verbosity >= 1:
      pr('Full_test_minors with pmax = {0}, nmax = {1}.\n'.format(pmax, nmax))
   counttests = 0
   countnonzero = 0
   tasks = []
   for p in range(pmax + 1):
      for n in range(nmax + 1):
         for d in range(n + 1):
            subs = list(Subsets(range(n), d))
            subs = [sorted(list(s)) for s in subs]
            tasks += [(p, n, xi, eta) for xi in subs for eta in subs]
   if verbosity >= 1:
      pr('Total number of minors to check: {0}.\n'.format(len(tasks)))
   for (p, n, xi, eta) in tasks:
      if not verify_minor(p, n, xi, eta, verbosity = verbosity):
         return False
      countnonzero += (our_minor_formula(p, n, xi, eta) != 0)
      counttests += 1
      if counttests % 10000 == 0 and verbosity >= 1:
         pr('{0} minors checked.\n'.format(counttests))
   if verbosity >= 1:
      pr('Total number of tests done: {0}.\n'.format(counttests))
      pr('Number of non-zero answers: {0}.\n'.format(countnonzero))
   return True

def verify_adjugate(p, n, verbosity = 0):
   if verbosity >= 2:
      pr('Verify_adjugate with p = {0}, n = {1} ... '.format(p, n))
   A = toeplitz_adjugate(p, n)
   correct = True
   for r in range(n):
      for s in range(n):
         a = A[r, s]
         b = our_adjugate_formula(p, n, r, s)
         correct = correct and (b == a)
         c = ale_adjugate_formula(p, n, r, s)
         correct = correct and (c == a)
         d = moc_adjugate_formula(p, n, r, s) if p >= 1 else a
         correct = correct and (d == a)
         e = tre_adjugate_formula(p, n, r, s)
         correct = correct and (e == a)
         if (not correct) and (verbosity >= 1):
            pr('Error with p = {0}, n = {1}, r = {2}, s = {3}:\n'.format(p, n, r, s))
            pr('Entry of the adjugate matrix: {0}\n'.format(a))
            pr('Our formula: {0}\n'.format(b))
            pr('Alexandersson formula: {0}\n'.format(c))
            pr('Moctezuma-Salazar formula: {0}\n'.format(d))
            pr('Trench formula: {0}\n'.format(e))
            return False
   if verbosity >= 2:
      pr('correct.\n')
   return True

def full_test_adjugates(pmax, nmax, verbosity = 0):
   tasks = []
   for p in range(pmax + 1):
      for n in range(1, nmax + 1):
         tasks += [(p, n)]
   if verbosity >= 1:
      pr('Full_test_adjugates with pmax = {0}, nmax = {1}.\n'.format(pmax, nmax))
      pr('Number of adjugate matrices to check: {0}.\n'.format(str(len(tasks))))
   counttests = 0
   for (p, n) in tasks:
      if not verify_adjugate(p, n, verbosity = verbosity):
         return False
      counttests += 1
   if verbosity >= 1:
      pr('Number of adjugate matrices checked: {0}.\n'.format(counttests))
   return True

def verify_eigenvector_property(p, n, verbosity = 0):
   # verify that T_n(a)v = (-1)^(p(n-1)) det(T_n(a)) e_1
   if verbosity >= 2:
      pr('Verify_eigenvector_property with p = {0}, n = {1} ... '.format(p, n))
   T = toeplitz_matrix(p, n)
   v = toeplitz_eigenvector(p, n)
   lhs = T * v
   D = ((-1) ** p) * schur([n] * p)
   rhs = matrix(schur, n, 1, [D] + [0] * (n - 1))
   correct = (lhs == rhs)
   if verbosity >= 2:
      pr('correct.\n' if correct else 'wrong.\n')
   if (not correct) and (verbosity >= 1):
      pr('Eigenvector is wrong for p = {0}, n = {1}.\n'.format(p, n))
   if (not correct and verbosity >= 1) or (verbosity >= 3):
      pr('T = \n{0}\n\n'.format(T))
      pr('v = \n{0}\n\n'.format(v))
      pr('T * v = \n{0}\n\n'.format(lhs))
      pr('det_with_a_certain_sign * basic_vector_1 = \n{0}\n\n'.format(rhs))
   return correct

def verify_eigenvector_expansion(p, n, w, verbosity = 0):
   if verbosity >= 2:
      pr('Verify_eigenvector_expansion with p = {0}, n = {1}, w = {2} ... '.format(p, n, w))
   v = toeplitz_eigenvector(p, n)
   vpol = matrix_from_schur_to_polynomial(v, w)
   wpol = toeplitz_eigenvector_via_geometric_progressions(p, n, w)
   correct = (wpol == vpol)
   if verbosity >= 2:
      pr('correct.\n' if correct else 'wrong.\n')
   if (not correct) and (verbosity >= 1):
      pr('Eigenvector is wrong for p = {0}, n = {1}, w = {2}.\n'.format(p, n, w))
   if (not correct and verbosity >= 1) or (verbosity >= 3):
      pr('Eigenvector from schur form:\n{0}\n'.format(vpol))
      pr('Eigenvector from geom progr:\n{0}\n'.format(wpol))
   return correct

def full_test_eigenvectors_property(pmax, nmax, verbosity = 0):
   tasks = [(p, n) for p in range(1, pmax + 1) for n in range(1, nmax + 1)]
   if verbosity >= 1:
      pr('Full_test_eigenvectors_property with pmax = {0}, nmax = {1}.\n'.format(pmax, nmax))
      pr('Number of matrices to check the eigenvector property: {0}.\n'.format(len(tasks)))
   counttests = 0
   for (p, n) in tasks:
      if not verify_eigenvector_property(p, n, verbosity = verbosity):
         return False
      counttests += 1
   if verbosity >= 1:
      pr('Number of matrices checked: {0}.\n'.format(counttests))
   return True

def full_test_eigenvectors_expansion(pmax, nmax, wmax, verbosity = 0):
   tasks = []
   for p in range(1, pmax + 1):
      for n in range(p, nmax + 1):
         for w in range(p, wmax + 1):
            tasks += [(p, n, w)]
   print('Number of matrices to check the eigenvector expansion: {0}'.format(str(len(tasks))))
   counttests = 0
   for (p, n, w) in tasks:
      if not verify_eigenvector_expansion(p, n, w, verbosity = verbosity):
         return False
      counttests += 1
   if verbosity >= 1:
      pr('Number of matrices checked: {0}.\n'.format(counttests))
   return True

def symbolic_tests_small():
   pr('{0}\n\n'.format(full_test_minors(3, 3, verbosity = 2)))
   pr('{0}\n\n'.format(full_test_adjugates(4, 4, verbosity = 2)))
   pr('{0}\n\n'.format(full_test_eigenvectors_property(4, 4, verbosity = 2)))
   pr('{0}\n\n'.format(full_test_eigenvectors_expansion(4, 4, 4, verbosity = 2)))

def symbolic_tests_huge():
   pr('{0}\n\n'.format(full_test_minors(8, 8, verbosity = 1)))
   pr('{0}\n\n'.format(full_test_adjugates(8, 8, verbosity = 1)))
   pr('{0}\n\n'.format(full_test_eigenvectors_property(8, 8, verbosity = 1)))
   pr('{0}\n\n'.format(full_test_eigenvectors_expansion(3, 8, 5, verbosity = 1)))

# symbolic_tests_small()

# A test of *one* minor can be very slow if, for example, n - d = 10.
# Therefore long_full_tests() works only with small values of p and n.
# In our computers, the following tests long various hours.
# The program uses about 15GB of RAM and about 5GB of swap memory.

# symbolic_tests_huge()

